function [x z nf ninf bestgen] = strawberry(x0, a, b, f, ngen, npop, nrmax, ns, population_strategy, output)

global esfdebug
global mosd_numberimproved
global npruned
global nsimilar
global strawberry_archive
global strawberry_fitness_method
global strawberry_diversity_tolerance
global strawberry_neighbourfn 
global strawberry_octave
global strawberry_output_population
mosd_numberimproved = 0;
npruned = 0;
nsimilar = 0;
stderr = 2;                     % for MATLAB compatibility
warning ('error', 'Octave:broadcast');
%% process arguments
if nargin<9, population_strategy = 4; end % population selection strategy
if nargin<8, ns = 100; end    % number of stable generations
if nargin<7, nrmax = 5; end   % number of runners
if nargin<6, npop = 100; end  % population size
if nargin<5, ngen = 1000; end % number of generations
if nargin<4
  disp('Error: need at least 4 arguments to Strawberry method: x0, a, b & f.')
  return
end
if nargin<10, 
  output = ngen/50;
end             % how often to output status of population
%% process global variables
if isempty (strawberry_archive)
  strawberry_archive = 0;
end
if strawberry_archive > 0
  fprintf(': pareto solutions are archived\n');
end
if isempty (strawberry_fitness_method)
  strawberry_fitness_method = 1;  %choose MOGA fitness for now
end
strawberry_diversity_tolerance = norm(b-a) / 1000;

%% determine the number of processors to use in parallel.  If the
%% variable has not been set, we use the system function to determine
%% how many processors there are
global strawberry_numberofprocessors
if isempty(strawberry_numberofprocessors)
  strawberry_numberofprocessors = 1 % not currently working for np>1, nproc();
end
if strawberry_numberofprocessors > 1
  fprintf (': using %d processors for parallel processing.\n', strawberry_numberofprocessors);
else
  fprintf (': not using any parallel processing on a single processor system.\n');
end
if strawberry_octave
  if strawberry_numberofprocessors > 1
    starttime = time;           % use wall clock time instead of CPU time
  else
    starttime = cputime;
  end
else
  tic;                        % start wall clock timer
  starttime = 0;              % for later use in output statistics
end
if isempty(strawberry_output_population)
  strawberry_output_population = 0
end
strawberry_initialise
 fprintf(': ngen=%d npop=%d nrmax=%d ns=%d population strategy=%d and tolerance=%f fitness method=%d\n', ngen, npop, nrmax, ns, population_strategy, strawberry_diversity_tolerance, strawberry_fitness_method)

constrained = 1;
try
  [z g] = f(x0);
catch err
  fprintf('Error identifier is %s\n', err.identifier);
  fprintf('Error message is %s\n', err.message);
  fprintf('\nThis could be due to the change in Strawberry that\n')
  fprintf('now requires the objective function to return a\n')
  fprintf('feasibility indication as well as the objective function\n')
  fprintf('value(s).\n\n')
  if (strcmp (err.identifier, 'Octave:undefined-function') || strcmp(err.identifier,'MATLAB:UndefinedFunction'))
    fprintf(': objective function is unconstrained.\n')
    constrained = 0;
  else
    %% not an error we expected to pass this on
    rethrow(err);
  end
end
if constrained
     fprintf(': objective function is constrained.\n');
end

[phi, nx, nz, nf, ninf] = strawberry_newpopulation (x0,a,b,npop,f,constrained);
pop = strawberry_sort (phi,nx,nz);
gen = 0;
best = [];                      %will be defined after fitness evaluation
bestgen = gen;
nfunctionevaluations = npop; % initial population size

if output > 0
  fprintf('In the following table, $n_f$ is number of feasible points,\n')
  fprintf('$n_i$ for infeasible, $n_p$ number in Pareto set\n');
  fprintf('$c$ best constraint value and $g_b$ best generation.\n');
  fprintf('| %9s |', 'gen')
  if nz > 1
    for i=1:nz, fprintf('  $z_{%02d}$ |', i); end
    if constrained; fprintf(' %9s |', 'c'); end
  else
    index = nx;
    if nx > 4
      index = 4;
    end
    for i=1:index; fprintf(' %6sx%02d |', '', i); end
    for i=1:nz; fprintf(' %6sy%02d |', '', i); end
    if constrained; fprintf(' %9s |', 'c'); end
  end
  fprintf(' %5s | %5s |', '$n_f$', '$n_i$')
  if nz > 1
    fprintf(' %5s |', '$n_p$')
  else
    fprintf(' %5s |', '$g_b$')
  end
  fprintf('\n');
  fprintf('|-\n');
end

while gen < ngen && (nz > 1 || gen < bestgen+ns)
  gen = gen + 1;

%% fprintf('npruned=%d nx=%d nz=%d\n', npruned, nx, nz)
pop = strawberry_prune (pop, nx, nz);
%% fprintf('npruned=%d nx=%d nz=%d\n', npruned, nx, nz)

%% evaluate the fitness of the population.  For multi-objective problems,
%% this has the side effect of returning the pareto front 
[N pareto] = strawberry_fitness (pop,nx,nz,size(pop,2));
%% populationstatistics('pareto',pareto,nx,nz);
%% keep track of best.  For multi-objective problems, this means keeping the pareto set
debugprint(1, 'best', best)
debugprint(1, 'pareto', pareto)
debugprint(1, 'pop(1)', pop(:,1))
debugprint(1, 'nx, nz', [nx, nz])
if nz < 2 ...
   && (isempty(best) ...
       || length(best) < 1 ...
       || (pareto(nx+1) < best(nx+1) ...
           && norm(abs(best(1:nx)-pareto(1:nx))) > 1e-6))
  best = pareto;
  bestgen = gen;
  %% zzz = zfit (best (1:nx))
  if ~output
    fprintf('\n... new best solution at generation %d, z(1)=%g x=', gen, best(nx+1));
    fprintf('%g ', best(1:nx));
    fprintf('\n');
  end
elseif nz>1
  %% update the best to be the set of non-dominated points from the union
  %% of the old best and the current population but only include old
  %% best if we wish to archive solutions
  if strawberry_archive > 0
    if isempty(best) || length(best) <= 1
      best = pareto;
    else
      union = [best, pareto];
      if length(union) < nx+nz+1
        best = [];
      else
        best = strawberry_prune(union(:,findpareto(union(nx+1:nx+nz,:))),nx,nz);
      end
    end
  else
    best = pareto;
  end
  debugprint (1, 'size of pareto: ', length(best))
end

%% populationstatistics('whole',pop,nx,nz);
%% populationstatistics('best',best,nx,nz);
if output > 0 && (0 == mod(gen,output) || 1 == gen)
  if strawberry_octave
    if strawberry_numberofprocessors > 1
      dtime = time-starttime;
    else
      dtime = cputime-starttime;
    end
  else
    dtime = toc;
  end
  fprintf(stderr, '\r');
  fprintf('| %9d |', gen);
  nfeas = sum(pop(end,:)<=0.0);
  ninfeas = sum(pop(end,:)>0);
  if nz > 1
    if ~ isempty(best) && length(best)>1
      %% fprintf('Size of best: %d %d\n',size(best))
      fprintf (' %9.3g |', min(best(nx+1:nx+nz,:),[],2))
      if constrained; fprintf(' %9.3g |', min(best(nx+nz+1,:))); end
    else
      fprintf(' %9.3g |', zeros(nz+constrained,1));
    end
    fprintf(' %5d | %5d | %5d |', nfeas, ninfeas, size(best,2));
  else
    if nx > 5
      indices = [1:4 nx+nz];
    else
      indices = 1:nx+nz;
    end
    fprintf (' %9.3g |', best(indices))
    if constrained; fprintf(' %9.3g |', best(nx+nz+1)); end
    fprintf(' %5d | %5d | %5d |', nfeas, ninfeas, bestgen);
  end
  fprintf('  \n');
else
  if nz > 1
    %% [fitness pareto] = moga_fitness (pop(1:nx,:), pop(nx+1:nx+nz,:));
    %% fprintf(stderr, '\r%30s %7d %3d/%4d %9.3g %9.3g %9.3g %9.3g', '', gen, size(best,2), length(pop), best(nx+1:nx+nz,1), best(nx+1:nx+nz,end))
    fprintf(stderr, '\r%30s %7d %3d/%4d ', '', gen, size(best,2), length(pop))
  else
    if constrained
      fprintf(stderr, '\r%30s %9d [%9d] %5d %13.6e %13.6e ', '', gen, bestgen, length(pop), best(nx+nz), best(nx+nz+1));
    else
      fprintf(stderr, '\r%30s %9d [%9d] %5d %8d %13.6e ', '', '', gen, bestgen, length(pop), nfunctionevaluations, best(nx+nz));
    end
  end
end

debugprint (1,'strawberry: fitness', N);
actualnpop = size(pop,2);
if actualnpop < size (pop,2)
  fprintf('Ummmm size of population %d is less than expected (%d)\n', size(1,pop), npop);
end

if strawberry_octave
  if strawberry_output_population > 0 ...
     && 0 == mod(gen,strawberry_output_population)
    filename = sprintf('%s-%s-%06d.txt', 'strawberry-population', ...
                       strftime('%Y%m%d-%H%M%S', localtime(time())), ...
                       gen);
    [sorted sortindices] = sort(best(nx+1,:));
    poptrans = best(:,sortindices)';
    save('-text',filename,'poptrans');
  end
end

%% generate all the new points and then evaluate them in parallel
n = 1;
newpop = [];
selected = [];
for p=1:npop                % we pick up to NPOP members to propagate
  s = strawberry_select (N);
  debugprint(2,'selected index', s)
  if s > 0
    selected(:,p) = pop(:,s);   % selected members remain for next generation; others die off
    debugprint(3, 'selected fitness and point: ', [N(s), pop(:,s)']);
    nr = strawberry_runners (N(s),nrmax); % number of runners
    debugprint(2,': number of runners', nr)
    for r=1:nr
      newpop{n} = strawberry_neighbourfn(pop(1:nx,s), a, b, N(s));
      debugprint(2, ': added new member to population: ', newpop{n})
      n = n + 1;
      %% end
    end
    %% set fitness so this member is not selected again
    N(s) = -1;
  end
end
debugprint(1,'Size of selected set: ', size(selected))
debugprint(1,'Size of newpop: ', size(newpop))
%% evaluate these in parallel using an appropriate number of processors
%%parf = @(x) if length(x)>nx, x', else [x, f(x), 0]', end
%%phi = parcellfun (strawberry_numberofprocessors, parf, newpop)';

if strawberry_numberofprocessors > 1
  if constrained
    phi = parcellfun (strawberry_numberofprocessors, @(x) [x; f(x); g(x)], newpop, 'VerboseLevel', 0);
  else
    phi = parcellfun (strawberry_numberofprocessors, @(x) [x; f(x); 0], newpop, 'VerboseLevel', 0);
  end
else                            %sequential evaluation
  n = length(newpop);
  x = newpop{1};
  if constrained
    [y g] = f(x);
  else
    y = f(x);
    g = 0;
  end
  phi = zeros(length(x)+length(y)+1,n);
  phi(:,1) = [x;y;g];
  for i=2:n
    x = newpop{i};
    nf = nf + 1;
    if constrained
      [y g] = f(x);
      if g > 0
        ninf = ninf + 1;
      end
    else
      y = f(x);
      g = 0;
    end
    phi(:,i) = [x;y;g];
  end
end
nfunctionevaluations = nfunctionevaluations + length(phi);
debugprint (1, ': phi before removal of infeasible solutions', phi)
ntotal = size(phi,2);

%% sort the population, which is composed of the original NPOP
%% best plus those created in this loop, using the fitness for
%% sorting.
%% we have three sets of solutions that we can combine to create a
%% new population:
%% 1. the best overall so far
%% 2. the members selected from the previous generation for propagation
%% 3. and the new propagated solutions
%% there are therefore 3! combinations although we always should
%% include the new solutions so really have 4 different options:
%% new alone, new with best, new with selected and new with
%% selected and best.

%% fprintf(stderr, 'size best=%d selected=%d phi=%d\n', size(best,2), size(selected,2), size(phi,2));
%% cater for when best array is empty
ps = population_strategy;
if ps > 2 && length(best) > (npop/2)
  %% do not do elitism if pareto set too large; kludge!
  ps = ps - 2;
end
switch (ps)
  case 1               % new members alone
    pop = phi;
  case 2               % new with selected
    pop = [selected, phi];
  case 3               % new with best
    pop = [best, phi];
  case 4               % new with best and selected
    pop = [best, selected, phi];
  otherwise
    fprintf('Error: population strategy %d not recognised.', population_strategy)
end

end
if gen < ngen
  fprintf('\n: strawberry terminating condition satisfied due to lack of improvement at gen=%d\n', gen);
end
if output
  fprintf('\n: strawberry at end, nf=%d ninf=%d best solution:\n', nf, ninf)
  disp(best');
end
%% ensure the best solution is returned
pop = [best, pop];
x = pop(1:nx,:);
z = pop(nx+1:end,:);
if strawberry_octave
  if strawberry_numberofprocessors > 1
    endtime = time;            % wall clock time
  else
    endtime = cputime;
  end
else
  endtime = toc;
end
 fprintf('\n: strawberry method elapsed time: %.1f with %d functions evaluated and %d solutions pruned with %d similar.\n', ...
       endtime-starttime, nf, npruned, nsimilar)
 fprintf('MOSD improvements: %d\n', mosd_numberimproved);
end
